//+------------------------------------------------------------------+
//|                                                        Trade.mqh |
//|           Copyright © 2023, Casper and Sharky, The Ghost Project |
//|                                   https://www.Coming_One_Day.com |
//+------------------------------------------------------------------+

#property copyright "Casper & Sharktoshi, The Ghost Project"
#property link      "Coming_One_Day.com"

#define MAX_RETRIES 5		// Max retries on error
#define RETRY_DELAY 3000	// Retry delay in ms

#ifndef INCLUDES
   	#define INCLUDES      
		#include <Trade\Trade.mqh>// DO NOT create a CTrade object in GhostTrade
		#include <errordescription.mqh>
		
		#include <Trade\Pending.mqh>
		CPending pendingOrders;
		
		#include <Ghost Project\TradingUtilities.mqh>
		CTradingUtilities TradingUtilities;
#endif 


//+------------------------------------------------------------------+
//| CTrade Class - Open, Close and Modify Orders                     |
//+------------------------------------------------------------------+

class CGhostTrade : public CTrade
{
	protected:
		bool m_EntryLevelHit;
		bool m_StopLossLevelHit;
		bool m_StopToScratchLevelHit;
		bool m_StopToTrailingLevelHit;
		bool m_TakeProfitLevelHit;
		bool m_TradeLevelsValidated;

		double m_NormalizedEntryLevel;
		double m_NormalizedStopLossLevel;
		double m_NormalizedTakeProfitLevel;
		double m_NormalizedTrailingStopLevel;
		double m_NormalizedStopToScratchLevel; 
		double m_NormalizedStopToTrailingLevel; 

		//TOOD: Why did I do the error codes this way? We already have a result object that we can use instead of making more variables. C-2023-09-08
		int m_TradeErrorCode;
		string m_TradeErrorString;		
	public:
		// TODO: We are going to need to create a default constructor so we can explicitly set certain variables such as m_TradeErrorString to be "Trade Level Validation Not Ran" or something like that. C-2023-08-26
		bool AdjustStopToStractch(); // TODO: Can we get away with this since stop to stractch will only ever happen 1 time per trade and we can just use the stored entry value? Also this should not need bullish/bearish functions since the functionality should be consistent across both. C-2023-09-22
		bool AdjustStopLossLevel(string pSymbol, double pStopLossLevel);
		bool AdjustBullishTrailingStopLevel(string pSymbol, double pTrailingStopLevel, double pCurrentBid, double pSymbolStopsLevel);
		bool AdjustBearishTrailingStopLevel(string pSymbol, double pTrailingStopLevel, double pCurrentAsk, double pSymbolStopsLevel);
		bool AdjustTrailingStopLevel(string pSymbol, double pTrailingStopLevel, double pCurrentBid, double pSymbolStopsLevel);
		bool BuyStopWithValidation(string pSymbol, int pNumContracts, double pEntryLevel, double pStopLossLevel, double pTakeProfitLevel, double pStopToScratchLevel, double pStopToTrailingLevel, datetime pExpiration = 0, string pComment = NULL);
		bool SellStopWithValidation(string pSymbol, int pNumContracts, double pEntryLevel, double pStopLossLevel, double pTakeProfitLevel, double pStopToScratchLevel, double pStopToTrailingLevel, datetime pExpiration = 0, string pComment = NULL);
		bool DeleteAllPendingOrders(string pSymbol);
		bool CheckForAndClearBullishPendingOrders(string pSymbol);
		bool BuyLimitWithValidation(string pSymbol, int pNumContracts, double pTakeProfitLevel, string pComment = NULL);
		bool SellLimitWithValidation(string pSymbol, int pNumContracts, double pTakeProfitLevel, string pComment = NULL);
		// Added by Casper and Sharky -- The Ghost Project
		/* C-2023-08-26 
			TODO: Create a GhostTrade class that extends CTrade and add these methods to it.
				- This is probably the proper way to do it but I didnt think we would need to do so many additions/changes to get what we want.
			
			****** Move the 'Note to Self' with the below code if it is relocated ******
			Note To Self: While it is true we could use the OrderInfo class instead of concentrating everythingh into the trade object, but
			it would require a similar amount of new functions to be written so instead Trade is not going to be our main point
			of reference for anything Trade related. 
			**********************************************************************************************************************	
		*/
		bool GetEntryLevelHit();
		bool GetStopLossLevelHit();
		bool GetStopToScratchLevelHit();
		bool GetStopToTrailingLevelHit();
		bool GetTakeProfitLevelHit();
		bool GetTradeLevelsValidated();

		double GetNormalizedEntryLevel();
		double GetNormalizedStopLossLevel();
		double GetNormalizedTakeProfitLevel();
		double GetNormalizedTrailingStopLevel();
		double GetNormalizedStopToScratchLevel(); 
		double GetNormalizedStopToTrailingLevel(); 
		
		string GetTradeErrorString();

		// TODO: Should all of these be made private/protected so that error codes can only be set within the class functions and not directly from the EA? C-2023-09-08
		void SetEntryLevelHit(bool pEntryLevelHit);
		void SetNormalizedEntryLevel(double pNormalizedEntryLevel);
		void SetNormalizedStopLossLevel(double pNormalizedStopLossLevel);
		void SetNormalizedTakeProfitLevel(double pNormalizedTakeProfitLevel);
		void SetNormalizedTrailingStopLevel(double pTrailingStopLevel);
		void SetNormalizedStopToScratchLevel(double pNormalizedStopToScratchLevel);  
		void SetNormalizedStopToTrailingLevel(double pNormalizedStopToTrailingLevel); 
		void SetStopLossLevelHit(bool pStopLossLevelHit);
		void SetStopToScratchLevelHit(bool pStopToScratch);
		void SetStopToTrailingLevelHit(bool pStopToTrailing);
		void SetTakeProfitLevelHit(bool pTakeProfitLevelHit);
		void SetTradeLevelsValidated(bool pTradeLevelsValidated);

		//TOOD: Why did I do the error codes this way? We already have a result object that we can use instead of making more variables. C-2023-09-08
		void SetTradeErrorCode(int pTradeErrorCode);
		void SetTradeErrorString(string pTradeErrorString);
};

bool CGhostTrade::GetEntryLevelHit(){
	return(m_EntryLevelHit);
}

bool CGhostTrade::GetStopLossLevelHit(){
	return(m_StopLossLevelHit);
}

bool CGhostTrade::GetStopToScratchLevelHit(){
	return(m_StopToScratchLevelHit);
}

bool CGhostTrade::GetStopToTrailingLevelHit(){
	return(m_StopToTrailingLevelHit);
}

bool CGhostTrade::GetTakeProfitLevelHit(){
	return(m_TakeProfitLevelHit);
}

bool CGhostTrade::GetTradeLevelsValidated(){
	return(m_TradeLevelsValidated);
}

double CGhostTrade::GetNormalizedEntryLevel(){
	return(m_NormalizedEntryLevel);
}

double CGhostTrade::GetNormalizedStopLossLevel(){
	return(m_NormalizedStopLossLevel);
}

double CGhostTrade::GetNormalizedTakeProfitLevel(){
	return(m_NormalizedTakeProfitLevel);
}

double CGhostTrade::GetNormalizedTrailingStopLevel(){ 
	return(m_NormalizedTrailingStopLevel); 
}

double CGhostTrade::GetNormalizedStopToScratchLevel(){ 
	return(m_NormalizedStopToScratchLevel); 
}

double CGhostTrade::GetNormalizedStopToTrailingLevel(){ 
	return(m_NormalizedStopToTrailingLevel); 
}

string CGhostTrade::GetTradeErrorString(){
	return(m_TradeErrorString);
}

void CGhostTrade::SetNormalizedEntryLevel(double pNormalizedEntryLevel){
	m_NormalizedEntryLevel = pNormalizedEntryLevel;
}

void CGhostTrade::SetNormalizedStopLossLevel(double pNormalizedStopLossLevel){
	m_NormalizedStopLossLevel = pNormalizedStopLossLevel;
}

void CGhostTrade::SetNormalizedTakeProfitLevel(double pNormalizedTakeProfitLevel){
	m_NormalizedTakeProfitLevel = pNormalizedTakeProfitLevel;
}

void CGhostTrade::SetNormalizedTrailingStopLevel(double pNormalizedTrailingStopLevel){ 
	m_NormalizedTrailingStopLevel = pNormalizedTrailingStopLevel; 
}

void CGhostTrade::SetNormalizedStopToScratchLevel(double pNormalizedStopToScratchLevel){ 
	m_NormalizedStopToScratchLevel = pNormalizedStopToScratchLevel; 
}

void CGhostTrade::SetNormalizedStopToTrailingLevel(double pNormalizedStopToTrailingLevel){ 
	m_NormalizedStopToTrailingLevel = pNormalizedStopToTrailingLevel; 
}

void CGhostTrade::SetStopLossLevelHit(bool pStopLossLevelHit){
	m_StopLossLevelHit = pStopLossLevelHit;
}

void CGhostTrade::SetStopToScratchLevelHit(bool pStopToScratchLevelHit){
	m_StopToScratchLevelHit = pStopToScratchLevelHit;
}

void CGhostTrade::SetStopToTrailingLevelHit(bool pStopToTrailingLevelHit){
	m_StopToTrailingLevelHit = pStopToTrailingLevelHit;
}

void CGhostTrade::SetTradeErrorCode(int pTradeErrorCode){
	m_TradeErrorCode = pTradeErrorCode;
}

void CGhostTrade::SetEntryLevelHit(bool pEntryLevelHit){
	m_EntryLevelHit = pEntryLevelHit;
}

void CGhostTrade::SetTakeProfitLevelHit(bool pTakeProfitLevelHit){
	m_TakeProfitLevelHit = pTakeProfitLevelHit;
}

void CGhostTrade::SetTradeLevelsValidated(bool pTradeLevelsValidated){
	m_TradeLevelsValidated = pTradeLevelsValidated;
}

void CGhostTrade::SetTradeErrorString(string pTradeErrorString){
	m_TradeErrorString = pTradeErrorString;
}

// TODO: Just realized we should not allow default of 0 on anything except Expiration and Comment as these levels cannot be 0
bool CGhostTrade::BuyStopWithValidation(string pSymbol, int pNumContracts, double pEntryLevel, double pStopLossLevel, double pTakeProfitLevel = 0, double pStopToScratchLevel = 0, double pStopToTrailingLevel = 0, datetime pExpiration = 0, string pComment = NULL){
	string symbol = pSymbol;
	int numContracts = pNumContracts;
	double entryLevel = pEntryLevel;
	double stopLossLevel = pStopLossLevel;
	double takeProfitLevel = pTakeProfitLevel;
	double stopToScratchLevel = pStopToScratchLevel;
	double stopToTrailingLevel = pStopToTrailingLevel;
	datetime expiration = pExpiration;
	string comment = pComment;
	
	bool success = false;
	bool validBullishTradeLevels = false;
	validBullishTradeLevels = TradingUtilities.ValidateBullishTradeLevels(symbol, numContracts, entryLevel, stopLossLevel, takeProfitLevel, stopToScratchLevel, stopToTrailingLevel);

	if(validBullishTradeLevels){
		
		string debugPrintString;
		string errorDescription;
	
		success = OpenPending(symbol, ORDER_TYPE_BUY_STOP, numContracts, entryLevel, stopLossLevel, 0.00, 0, expiration, comment); // Does not open with a take profit level. This is added after the order is placed due to the way the IWBDT setup works and our requirement of having a Trailing Stop Loss. C-2023-09-17

		int checkCode = CheckReturnCode(result.retcode);
		if(checkCode == CHECK_RETCODE_OK || success){
			errorDescription = TradeServerReturnCodeDescription(result.retcode);
			TradingUtilities.DebugPrint("Ret Code: " + checkCode + "Error Description: " + errorDescription);	
			
			SetNormalizedEntryLevel(entryLevel);
			SetNormalizedStopLossLevel(stopLossLevel);
			SetNormalizedTakeProfitLevel(takeProfitLevel);
			SetNormalizedStopToScratchLevel(stopToScratchLevel);
			SetNormalizedStopToTrailingLevel(stopToTrailingLevel);
			SetTradeLevelsValidated(true);
			SetEntryLevelHit(false);
			SetStopLossLevelHit(false);
			SetTakeProfitLevelHit(false);
			SetStopToScratchLevelHit(false);
			SetStopToTrailingLevelHit(false);
		}else{
			errorDescription = TradeServerReturnCodeDescription(result.retcode);
			debugPrintString = "Bullish Trade Order Placement Error: " + checkCode + " Error Description: " + errorDescription + " Detected. Checking for pending orders and retrying...";
			TradingUtilities.DebugPrint(debugPrintString);	
			success = false; // probably redundant C-28-09-23
			SetTradeLevelsValidated(false); 

			if(checkCode == CHECK_RETCODE_ERROR){
				bool pendingOrdersDetectedAndCleared = CheckForAndClearBullishPendingOrders(_Symbol);
				SetTradeErrorString(errorDescription);
				TradingUtilities.DebugPrint(debugPrintString);								
			}else{
				bool pendingOrdersDetectedAndCleared = CheckForAndClearBullishPendingOrders(_Symbol);
				string errorDescription = TradeServerReturnCodeDescription(result.retcode);
				SetTradeErrorString(errorDescription);
			}
		}
		
	}else{
		success = false;
		string debugPrintString = "Bullish Trade Levels Failed To Validate";
		TradingUtilities.DebugPrint(debugPrintString);
		TradingUtilities.DebugPrint(GetTradeErrorString());
		//TODO: Need to add detailed error handling/logging here to determine which level was not valid. C-2023-08-26
	}

	return(success);
}

bool CGhostTrade::SellStopWithValidation(string pSymbol, int pNumContracts, double pEntryLevel, double pStopLossLevel, double pTakeProfitLevel = 0, double pStopToScratchLevel = 0, double pStopToTrailingLevel = 0, datetime pExpiration = 0, string pComment = NULL){
	string symbol = pSymbol;
	int numContracts = pNumContracts;
	double entryLevel = pEntryLevel;
	double stopLossLevel = pStopLossLevel;
	double takeProfitLevel = pTakeProfitLevel;
	double stopToScratchLevel = pStopToScratchLevel;
	double stopToTrailingLevel = pStopToTrailingLevel;
	datetime expiration = pExpiration;
	string comment = pComment;
	
	bool success = false;
	bool validBearishTradeLevels = false;
	validBearishTradeLevels = TradingUtilities.ValidateBearishTradeLevels(symbol, numContracts, entryLevel, stopLossLevel, takeProfitLevel, stopToScratchLevel, stopToTrailingLevel);

	if(validBearishTradeLevels){
		
		string debugPrintString;
		string errorDescription;
	
		// TODO: We should probably be setting a TP right away and then modifying it to be 0 after the order is placed instead of essentially making it so that you can never set a TP on order open and it would require a seperate call for all setups. C-2023-09-28
		success = OpenPending(symbol, ORDER_TYPE_SELL_STOP, numContracts, entryLevel, stopLossLevel, 0.00, 0, expiration, comment); // Does not open with a take profit level. This is added after the order is placed due to the way the IWBDT setup works and our requirement of having a Trailing Stop Loss. C-2023-09-17

		int checkCode = CheckReturnCode(result.retcode);
		if(checkCode == CHECK_RETCODE_OK || success){
			errorDescription = TradeServerReturnCodeDescription(result.retcode);
			
			SetNormalizedEntryLevel(entryLevel);
			SetNormalizedStopLossLevel(stopLossLevel);
			SetNormalizedTakeProfitLevel(takeProfitLevel);
			SetNormalizedStopToScratchLevel(stopToScratchLevel);
			SetNormalizedStopToTrailingLevel(stopToTrailingLevel);
			SetTradeLevelsValidated(true);
			SetEntryLevelHit(false);
			SetStopLossLevelHit(false);
			SetTakeProfitLevelHit(false);
			SetStopToScratchLevelHit(false);
			SetStopToTrailingLevelHit(false);
		}else{
			errorDescription = TradeServerReturnCodeDescription(result.retcode);
			success = false; // TODO: probably redundant C-28-09-23
			SetTradeLevelsValidated(false); 

			if(checkCode == CHECK_RETCODE_ERROR){
				bool pendingOrdersDetectedAndCleared = CheckForAndClearBullishPendingOrders(_Symbol);
				SetTradeErrorString(errorDescription);
				TradingUtilities.DebugPrint(debugPrintString);								
			}else{
				bool pendingOrdersDetectedAndCleared = CheckForAndClearBullishPendingOrders(_Symbol);	
				string errorDescription = TradeServerReturnCodeDescription(result.retcode);	
				SetTradeErrorString(errorDescription);
			}
		}
		
	}else{
		success = false;
		string debugPrintString = "Bearish Trade Levels Failed To Validate";
		TradingUtilities.DebugPrint(debugPrintString);
		TradingUtilities.DebugPrint(GetTradeErrorString());
		//TODO: Need to add detailed error handling/logging here to determine which level was not valid. C-2023-08-26
	}

	return(success);
}

bool CGhostTrade::DeleteAllPendingOrders(string pSymbol){
	bool success = false;
	string symbol = pSymbol;
	
	int totalPendingOrders = pendingOrders.TotalPending(symbol);
	ulong ticketOrderNumbers[];
	
	pendingOrders.GetTickets(symbol, ticketOrderNumbers);
	for (int i = 0; i < totalPendingOrders; i++){
		Print("Deleting Ticket Order Number: " + IntegerToString(ticketOrderNumbers[i]));
		Delete(ticketOrderNumbers[i]);
	}

	//Make sure there are 0 orders remaining
	totalPendingOrders = pendingOrders.TotalPending(symbol);
	if (totalPendingOrders == 0){
		return true;
	}else{
		return false;
	}
} 

bool CGhostTrade::CheckForAndClearBullishPendingOrders(string pSymbol){
	string symbol = pSymbol;
	if(OrdersTotal() > 0){
		bool pendingOrdersDeleted = DeleteAllPendingOrders(symbol);
		if(pendingOrdersDeleted){
			return true;
		}else{
			return false;
		}
	}else{
		return true;
	}
}

bool CGhostTrade::BuyLimitWithValidation(string pSymbol, int pNumContracts, double pTakeProfitLevel, string pComment = NULL){
	string symbol = pSymbol;
	int numContracts = pNumContracts;
	double takeProfitLevel = pTakeProfitLevel;
	string comment = pComment;
	bool success = false;
	double normalizedTakeProfitLevel = GetNormalizedTakeProfitLevel();

	if(takeProfitLevel != normalizedTakeProfitLevel){
		TradingUtilities.DebugPrint("CRITICAL ERROR: Take Profit Level Does Not Match."); // If this error appears then we really need to consider whether or not we should rely only on the passed pTakeProfitLevel or the stored m_TakeProfitLevel C-2023-09-17
		return false;
	}

	bool validTakeProfitLevel = TradingUtilities.ValidateTakeProfitLevel(symbol, normalizedTakeProfitLevel); // should this be the passed TP instead of the stored one we are getting? Unsure at the moment. Future Craig will figure it out if this ever comes up as an issue. C-2023-09-17
	if(validTakeProfitLevel){
		success = BuyLimit(symbol, 1, normalizedTakeProfitLevel);

		if(!success){ // prob need to move this down and return out of the else if instead of inside here. C-2023-09-17
			return false;
		}
	}

	int checkCode = CheckReturnCode(result.retcode);

	if(checkCode == CHECK_RETCODE_OK){
		// not sure what to put here since success is set above from the result of SellLimit. C-2023-09-17
	}else if(checkCode == CHECK_RETCODE_ERROR){
		string errorDescription = TradeServerReturnCodeDescription(result.retcode);
		success = false;
		SetTradeErrorString(errorDescription);
		string debugPrintString = "Bearish Take Profit Failed: Error " + result.retcode + " - " + errorDescription;
		TradingUtilities.DebugPrint(debugPrintString);
	}else{
		string errorDescription = "Unknown Bearish Take Profit Error Detected...";
		// Do we want to exit trade/clean up orders/anything here if this error occurs? C-2023-08-28

		SetTradeErrorString(errorDescription);
		TradingUtilities.DebugPrint(errorDescription);
	}
	
	return(success);
}

bool CGhostTrade::SellLimitWithValidation(string pSymbol, int pNumContracts, double pTakeProfitLevel, string pComment = NULL){
	string symbol = pSymbol;
	int numContracts = pNumContracts;
	double takeProfitLevel = pTakeProfitLevel;
	string comment = pComment;
	bool success = false;
	double normalizedTakeProfitLevel = GetNormalizedTakeProfitLevel();

	if(takeProfitLevel != normalizedTakeProfitLevel){
		TradingUtilities.DebugPrint("CRITICAL ERROR: Take Profit Level Does Not Match."); // If this error appears then we really need to consider whether or not we should rely only on the passed pTakeProfitLevel or the stored m_TakeProfitLevel C-2023-09-17
		return false;
	}

	bool validTakeProfitLevel = TradingUtilities.ValidateTakeProfitLevel(symbol, normalizedTakeProfitLevel); // should this be the passed TP instead of the stored one we are getting? Unsure at the moment. Future Craig will figure it out if this ever comes up as an issue. C-2023-09-17
	if(validTakeProfitLevel){
		success = SellLimit(symbol, numContracts, normalizedTakeProfitLevel);

		if(!success){ // prob need to move this down and return out of the else if instead of inside here. C-2023-09-17
			return false;
		}
	}

	int checkCode = CheckReturnCode(result.retcode);

	if(checkCode == CHECK_RETCODE_OK){
		// not sure what to put here since success is set above from the result of SellLimit. C-2023-09-17
	}else if(checkCode == CHECK_RETCODE_ERROR){
		string errorDescription = TradeServerReturnCodeDescription(result.retcode);
		success = false;
		SetTradeErrorString(errorDescription);
		string debugPrintString = "Bullish Take Profit Failed: Error " + result.retcode + " - " + errorDescription;
		TradingUtilities.DebugPrint(debugPrintString);
	}else{
		string errorDescription = "Unknown Bullish Take Profit Error Detected...";
		// Do we want to exit trade/clean up orders/anything here if this error occurs? C-2023-08-28

		SetTradeErrorString(errorDescription);
		TradingUtilities.DebugPrint(errorDescription);
	}
	
	return(success);
}

// TODO: Can this be combined with AdjustStopLossLevel below? C-2023-09-22
bool CGhostTrade::AdjustStopToStractch(){
	bool tradeLevelsValidated = GetTradeLevelsValidated();
	bool success = false;
	string debugPrintString;
	string errorDescription;
	Print("@#@#@#* Stop To Scratch: " + tradeLevelsValidated);
	// Print("Stop To Scratch: Stop to scratchLevelHit: " + stopToScratchLevelHit + " stopToScratchLevelGreaterThanZero: " + stopToScratchLevelGreaterThanZero + "stopLossHasNotExceededEntryLevel" + stopLossHasNotExceededEntryLevel + " Bar.High(0): " + Bar.High(0) + " normalizedStopToScratchLevel: " + normalizedStopToScratchLevel);
	if(tradeLevelsValidated){
		success = AdjustStopLossLevel(_Symbol, GetNormalizedEntryLevel());
		SetStopToScratchLevelHit(true);
		// double normalizedStopLossLevel = GetNormalizedEntryLevel();
		// success = ModifyPosition(_Symbol, normalizedStopLossLevel);

		// int checkCode = CheckReturnCode(result.retcode);
		// if(checkCode == CHECK_RETCODE_OK){
		// 	SetNormalizedStopLossLevel(normalizedStopLossLevel);
		// 	SetStopToScratchLevelHit(true);
		// }else if(checkCode == CHECK_RETCODE_ERROR){
		// 	errorDescription = TradeServerReturnCodeDescription(result.retcode);
		// 	SetTradeErrorString(errorDescription);
		// 	debugPrintString = "Error: Adjust Stop To Scratch Position Failed: " + result.retcode + " - " + errorDescription;
		// 	TradingUtilities.DebugPrint(debugPrintString);
		// }else{
		// 	errorDescription = "Error: Unknown Adjust Stop To Scratch Issue Detected. Retrying...";
		// 	success = false;
		// 	SetTradeErrorString(errorDescription);
		// 	TradingUtilities.DebugPrint(errorDescription);
		// }
	}else{
		errorDescription = "Error: Trade Levels Not Validated. Adjust Stop To Scratch Failed.";
		SetTradeErrorString(errorDescription);
		TradingUtilities.DebugPrint(errorDescription);
		success = false;
	}

	return success;
}

// AdjustStopLossLevel is not a "secure" function. It will not validate any information passed to it other than to normalize it. (normalization may be removed in the future. Not sure ) - C-2023-09-24
bool CGhostTrade::AdjustStopLossLevel(string pSymbol, double pStopLossLevel){
	double normalizedStopLossLevel = TradingUtilities.NormalizePriceUsingTickSize(pSymbol, pStopLossLevel);
	bool success = ModifyPosition(_Symbol, normalizedStopLossLevel);
	int checkCode = CheckReturnCode(result.retcode);
	string debugPrintString;
	string errorDescription;
	
	if(checkCode == CHECK_RETCODE_OK || success){
		SetNormalizedStopLossLevel(normalizedStopLossLevel);
	}else if(checkCode == CHECK_RETCODE_ERROR){
		errorDescription = TradeServerReturnCodeDescription(result.retcode);
		SetTradeErrorString(errorDescription);
		debugPrintString = "Error: Adjust Stop Loss Level Failed: " + result.retcode + " - " + errorDescription;
		TradingUtilities.DebugPrint(debugPrintString);
	}else{
		errorDescription = "Error: Unknown Adjust Stop Loss Level Issue Detected.";
		SetTradeErrorString(errorDescription);
		TradingUtilities.DebugPrint(errorDescription);
	}
	return success;
}

bool CGhostTrade::AdjustBullishTrailingStopLevel(string pSymbol, double pTrailingStopLevel, double pCurrentBid, double pSymbolStopsLevel){
	bool success = false;
	double previousNormalizedTrailingStopLevel = GetNormalizedStopLossLevel();
	double normalizedTrailingStopLevel = TradingUtilities.CalculateBullishStopLossLevel(pTrailingStopLevel);
	
	if(!(previousNormalizedTrailingStopLevel == normalizedTrailingStopLevel)){
		if(normalizedTrailingStopLevel < (pCurrentBid - pSymbolStopsLevel)){
			success = AdjustTrailingStopLevel(pSymbol, normalizedTrailingStopLevel, pCurrentBid, pSymbolStopsLevel);
		}
	}
	return success;	
}

bool CGhostTrade::AdjustBearishTrailingStopLevel(string pSymbol, double pTrailingStopLevel, double pCurrentAsk, double pSymbolStopsLevel){
	bool success = false;
	double previousNormalizedTrailingStopLevel = GetNormalizedStopLossLevel();
	double normalizedTrailingStopLevel = TradingUtilities.CalculateBearishStopLossLevel(pTrailingStopLevel);
	
	if(!(previousNormalizedTrailingStopLevel == normalizedTrailingStopLevel)){
		if(normalizedTrailingStopLevel > (pCurrentAsk + pSymbolStopsLevel)){
			success = AdjustTrailingStopLevel(pSymbol, normalizedTrailingStopLevel, pCurrentAsk, pSymbolStopsLevel);
		}
	}
	return success;
}

bool CGhostTrade::AdjustTrailingStopLevel(string pSymbol, double pTrailingStopLevel, double pCurrentBid, double pSymbolStopsLevel){
	int checkCode = 0;
	bool success = false;
	string debugPrintString;
	string errorDescription;

	success = AdjustStopLossLevel(_Symbol, pTrailingStopLevel);
	checkCode = CheckReturnCode(result.retcode);
	if(checkCode == CHECK_RETCODE_OK || success){
		SetNormalizedStopLossLevel(pTrailingStopLevel);
		SetStopToTrailingLevelHit(true);
	}else if(checkCode == CHECK_RETCODE_ERROR){
		errorDescription = TradeServerReturnCodeDescription(result.retcode);
		SetTradeErrorString(errorDescription);
		debugPrintString = "Error: Adjust Trailing Stop Failed: " + "Normalized Trailing Stop: " + pTrailingStopLevel + result.retcode + " - " + errorDescription;
		TradingUtilities.DebugPrint(debugPrintString);
		success = false;
	}else{
		errorDescription = "Critical Error: Unknown AdjustTrailingStopLevel Error Detected.";
		SetTradeErrorString(errorDescription);
		TradingUtilities.DebugPrint(errorDescription);
		success = false;
	}
	return success;
}